
import argparse
from pathlib import Path
import torch
from tqdm import tqdm
import math

from splat_viewer.gaussians.data_types import Gaussians
from splat_viewer.gaussians.workspace import load_workspace

from taichi_splatting import RasterConfig, map_to_tiles, pad_to_tile, perspective, tile_mapper

import taichi as ti
from splat_viewer.renderer.taichi_splatting import to_camera_params


def overlap_stats(overlap_counts, tile_counts,  max_hist = 8):
    large_thresh = (overlap_counts.std() * 3).item()
    very_large = (overlap_counts > large_thresh).sum()

    hist = torch.histc(overlap_counts.clamp(0, max_hist + 1), bins=max_hist + 1, min=0, max=max_hist + 1)

    n = overlap_counts.shape[0]
    print(f"Overlap mean: {overlap_counts.mean():.2f} very large (> {large_thresh:.2f}) %: {100.0 * very_large / n:.2f}")
    print(f"Overlap histogram %: {100 * hist[:max_hist] / n} > 10: {100 * hist[max_hist] / n}")


    print(f"Mean tile count: {tile_counts.mean():.2f}, max tile count: {tile_counts.max()}")


def main():

  torch.set_printoptions(precision=5, sci_mode=False, linewidth=120)

  parser = argparse.ArgumentParser()

  parser.add_argument("model_path", type=Path, help="workspace folder containing cameras.json, input.ply and point_cloud folder with .ply models")
  parser.add_argument("--device", type=str, default="cuda:0", help="torch device to use")
  parser.add_argument("--model", type=str, default=None, help="model to load")

  parser.add_argument("--tile_size", type=int, default=16, help="tile size for rasterizer")
  parser.add_argument("--image_size", type=int, default=None, help="resize longest edge of camera image sizes")
  parser.add_argument("--no_tight_culling", action="store_true", help="disable tight culling")
  parser.add_argument("--debug", action="store_true", help="enable taichi debug mode")

  args = parser.parse_args()

  ti.init(arch=ti.cuda, offline_cache=True, log_level=ti.INFO,
          debug=args.debug, device_memory_GB=0.1)


  workspace = load_workspace(args.model_path)
  if args.model is None:
      args.model = workspace.latest_iteration()
    
  gaussians:Gaussians = workspace.load_model(args.model).to(args.device)
  gaussians = gaussians.sorted()


  print(f"Using {args.model_path} with {gaussians.batch_size[0]} points")

  image_sizes = set([tuple(camera.image_size) for camera in workspace.cameras])
  print(f"Cameras: {len(workspace.cameras)}, Image sizes: {image_sizes}")
    
  config = RasterConfig(
    tile_size=args.tile_size, 
    tight_culling=not args.no_tight_culling)

  packed = gaussians.packed()

  overlaps = []
  tile_counts = []

  for camera in tqdm(workspace.cameras):
    camera_params = to_camera_params(camera, device=args.device)

    mask = perspective.frustum_culling(packed, camera_params, margin_pixels=50)
    gaussians2d, depth = perspective.project_to_image(packed[mask], camera_params)
      
    overlap_to_point, tile_ranges = map_to_tiles(
      gaussians2d, depth, camera_params.image_size, config)
    
    image_size = pad_to_tile(camera.image_size, config.tile_size)

    mapper = tile_mapper.tile_mapper(config)
    overlap_offsets, total_overlap = mapper.generate_tile_overlaps(
      gaussians2d, image_size)
  
    cum_overlap_counts = torch.cat([overlap_offsets.cpu(), torch.tensor([total_overlap])])

    
    overlaps.append((cum_overlap_counts[1:] - cum_overlap_counts[:-1]).float())
    tile_counts.append((tile_ranges[:, 1] - tile_ranges[:, 0]).float())


  overlap_stats(torch.cat(overlaps), torch.cat(tile_counts))
  max_counts = [torch.max(overlap) for overlap in overlaps]
  print(f"Max overlap (mean): {sum(max_counts) / len(max_counts):.2f}")
     

if __name__ == "__main__":
  main()  




from .loading import from_pcd, to_pcd, read_gaussians, write_gaussians

from .data_types import Rendering, Gaussians

__all__ = [
    "Gaussians",
    "from_pcd",
    "to_pcd",
    "read_gaussians",
    "write_gaussians",
    "Rendering"
]


from dataclasses import replace
from typing import Tuple

from PySide6 import QtGui, QtCore, QtWidgets
from PySide6.QtCore import Qt, QEvent
from beartype import beartype
import cv2

import math

from pathlib import Path

import numpy as np
import torch
from splat_viewer.camera.fov import FOVCamera

from splat_viewer.gaussians.workspace import Workspace
from splat_viewer.gaussians import Gaussians
from splat_viewer.viewer.interactions.animate import  animate_to_loop
from splat_viewer.viewer.interaction import Interaction
from splat_viewer.viewer.renderer import WorkspaceRenderer

    
from .interactions.fly_control import FlyControl
from .scene_camera import SceneCamera
from .settings import Settings, ViewMode



class SceneWidget(QtWidgets.QWidget):
  def __init__(self, settings:Settings = Settings(), renderer=None):
    super(SceneWidget, self).__init__()

    SceneWidget.instance = self

    self.camera_state = Interaction()
    self.interaction = Interaction()

    self.camera = SceneCamera()
    self.settings = settings
    self.renderer = renderer

    self.setFocusPolicy(Qt.StrongFocus)
    self.setMouseTracking(True)

    
    self.cursor_pos = (0, 0)
    self.modifiers = Qt.NoModifier
    self.keys_down = set()

    self.dirty = True

    self.timer = QtCore.QTimer(self)
    self.timer.timeout.connect(self.update)
    self.timer.start(1000 / Settings.update_rate)



  def update_setting(self, **kwargs):
    self.settings = replace(self.settings, **kwargs)
    self.dirty = True


  def load_workspace(self, workspace:Workspace, gaussians:Gaussians):
    self.workspace = workspace
    
    self.workspace_renderer = WorkspaceRenderer(workspace, gaussians.to(device=self.settings.device), self.renderer)
    self.keypoints = self.read_keypoints()

    self.set_camera_index(0)
    self.camera_state = FlyControl()
    


  @property
  def camera_path_file(self):
    return self.workspace.model_path / "camera_path.npy"

  def write_keypoints(self):
    np.save(self.camera_path_file, np.array(self.keypoints))
    print(f"Saved {len(self.keypoints)} keypoints to {self.camera_path_file}")

  def read_keypoints(self):
    if self.camera_path_file.exists():
      kp = list(np.load(self.camera_path_file))
      print(f"Loaded {len(kp)} keypoints from {self.camera_path_file}")
      return kp

    return []



  def set_camera_index(self, index:int):
    self.camera_state.transition(None)

    camera = self.workspace.cameras[index]
    print('Showing view from camera', camera.image_name)
    self.zoom = 1.0

    self.camera.set_camera(camera)
    self.camera_index = index
    self.dirty = True


  @property
  def image_size(self):
    w, h = self.size().width(), self.size().height()

    return w, h
  
  def sizeHint(self):
    return QtCore.QSize(1024, 768)

  def event(self, event: QEvent):

    if (self.interaction.trigger_event(event) or 
        self.camera_state.trigger_event(event)):
      return True
      
    return super(SceneWidget, self).event(event)

    

  def keyReleaseEvent(self, event: QtGui.QKeyEvent) -> bool:
    self.modifiers = event.modifiers()
    self.keys_down.discard(event.key())


    return super().keyPressEvent(event)
  
  def focusOutEvent(self, event: QtGui.QFocusEvent):
    self.keys_down.clear()
    return super().focusOutEvent(event)    
  
  def keyPressEvent(self, event: QtGui.QKeyEvent) -> bool:
    self.modifiers = event.modifiers()
    self.keys_down.add(event.key())

    view_modes = {
      Qt.Key_1 : ViewMode.Normal,
      Qt.Key_2 : ViewMode.Points,
      Qt.Key_3 : ViewMode.Depth,

    }

    enable_disable = {
      Qt.Key_0 : 'cropped',
      Qt.Key_9 : 'initial_points',
      Qt.Key_8 : 'cameras',
    }

    if event.key() == Qt.Key_Print:
      self.save_snapshot()
      return True
  
    elif event.key() == Qt.Key_BraceLeft:
      self.set_camera_index((self.camera_index - 1) % len(self.workspace.cameras))
      return True
    elif event.key() == Qt.Key_BraceRight:
      self.set_camera_index((self.camera_index + 1) % len(self.workspace.cameras))
      return True
    

    elif event.key() == Qt.Key_Equal: 
      self.camera.zoom(self.settings.zoom_discrete)
      self.dirty = True
      return True
    elif event.key() == Qt.Key_Minus:
      self.camera.zoom(1/self.settings.zoom_discrete)
      self.dirty = True

      return True

    elif event.key() in enable_disable:
      k = enable_disable[event.key()]
      update = {k: not getattr(self.settings.show, k)}
      
      self.update_setting(show = replace(self.settings.show, **update))
      return True
    
    elif event.key() in view_modes:
      k = view_modes[event.key()]
      self.update_setting(view_mode = k)
      return True
    
    
    elif event.key() == Qt.Key_Space:
      self.keypoints.append(self.camera.view_matrix)

    if event.key() == Qt.Key_Space and event.modifiers() & Qt.ControlModifier:  
      self.write_keypoints()

    elif event.key() == Qt.Key_Return:
      if event.modifiers() & Qt.ShiftModifier:
        if self.window().isFullScreen():
          self.window().showNormal()
        else:
          self.window().showFullScreen()
      elif len(self.keypoints) > 0:
        animate_to_loop(self.camera_state, 
                        self.camera.view_matrix, self.keypoints)

    return super().keyPressEvent(event)
  

  def update(self):
    self.camera_state._update(1 / self.settings.update_rate)
    self.repaint()

  def resizeEvent(self, event: QtGui.QResizeEvent):
    self.dirty = True
    return super().resizeEvent(event)

  def mouseMoveEvent(self, event: QtGui.QMouseEvent):
    p = event.localPos()
    self.cursor_pos = (p.x(), p.y())
    return super().mouseMoveEvent(event)

  def current_point_3d(self) -> np.ndarray:
    return self.lookup_point_3d(self.cursor_pos)
    
  @property
  def rendering(self):
    if self.workspace_renderer.rendering is None:
      raise ValueError("No depth render available")

    return self.workspace_renderer.rendering

  def unproject_point(self, p:np.ndarray, depth:float) -> np.ndarray:
    render = self.rendering
    camera:FOVCamera = render.camera
    scene_point =  camera.unproject_pixel(*p, depth)
    return np.array([scene_point], dtype=np.float32)
  

  def unproject_radius(self, p:np.ndarray, depth:float, radius:float
                       ) -> Tuple[np.ndarray, float]:
    p = np.array(p)
    
    p1 = self.unproject_point(p, depth)
    p2 = self.unproject_point(p + np.array([radius, 0]), depth)

    return p1, np.linalg.norm(p2 - p1)


  def lookup_depth(self, p:Tuple[int, int]) -> np.ndarray:
    render = self.rendering
    
    p = np.round(p).astype(np.int32)
    x = np.clip(p[0], 0, render.depth.shape[1] - 1)
    y = np.clip(p[1], 0, render.depth.shape[0] - 1)

    return render.depth[y, x].item()

  def from_numpy(self, a:np.ndarray):
    return torch.from_numpy(a).to(device=self.settings.device)

  @beartype
  def lookup_depths(self, p:np.ndarray) -> np.ndarray:
    render = self.rendering
    p = np.round(p).astype(np.int32)

    x = np.clip(p[:, 0], 0, render.depth.shape[1] - 1)
    y = np.clip(p[:, 1], 0, render.depth.shape[0] - 1)

    x, y = self.from_numpy(x), self.from_numpy(y)
    return render.depth[y, x].cpu().numpy()

  @beartype
  def test_depths(self, p:np.ndarray, depth:np.ndarray, tol=0.98) -> np.ndarray:

    return ((depth * tol <= self.lookup_depths(p)) & 
            (p[:, 0] >= 0) & (p[:, 0] < self.image_size[0] - 1) & 
            (p[:, 1] >= 0) & (p[:, 1] < self.image_size[1] - 1))


  @property
  def depth_map(self) -> torch.Tensor:
    render = self.rendering
    return render.depth

  def lookup_point_3d(self, p:np.ndarray) -> np.ndarray:
    render = self.rendering
    scene_point =  render.camera.unproject_pixel(*p, self.lookup_depth(p))
    return np.array([scene_point], dtype=np.float32)


  def render_camera(self) -> FOVCamera:
    return self.camera.resized(self.image_size)
  
  def render(self):
    camera = self.render_camera()

    self.view_image = np.ascontiguousarray(
      self.workspace_renderer.render(camera, self.settings))
    
    self.dirty = False
    return self.view_image

      
  def paintEvent(self, event: QtGui.QPaintEvent):
    with QtGui.QPainter(self) as painter:
      dirty = self.dirty
      if dirty:
        self.render()

      painter.drawImage(0, 0, QtGui.QImage(self.view_image, 
                  self.view_image.shape[1], self.view_image.shape[0],  
                  QtGui.QImage.Format_RGB888))
      
            
  def snapshot_file(self):
    pictures = Path.home() / "Pictures"
    filename = pictures / "snapshot.jpg"

    i = 0
    while filename.exists():
      i += 1
      filename = pictures / f"snapshot_{i}.jpg"

    return filename


  def render_tiled(self, camera:FOVCamera):
    tile_size = self.settings.snapshot_tile
    nw, nh = [int(math.ceil(x / tile_size)) 
              for x in camera.image_size]
    
    full_image = np.zeros((nh * tile_size, nw * tile_size, 3), dtype=np.uint8)
    
    for x in range(0, nw):
      for y in range(0, nh):
        tile_camera = camera.crop(np.array([x * tile_size, y * tile_size]), 
                             np.array([tile_size, tile_size]))
        
        image = self.workspace_renderer.render(tile_camera, self.settings)
        tile = full_image[y * tile_size:(y + 1) * tile_size, 
                          x * tile_size:(x + 1) * tile_size, :] 
        
        print(tile.shape, image.shape, x, y)
        tile[:] = image
        
    return full_image[:camera.image_size[1], :camera.image_size[0]]

  def save_snapshot(self):
    camera = self.camera.resized(self.settings.snapshot_size)
    filename = self.snapshot_file()

    w, h = camera.image_size
    print(f"Rendering snapshot ({w}x{h})...")
    print(camera)

    image = self.render_tiled(camera)

    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite(str(filename), image, [cv2.IMWRITE_JPEG_QUALITY, 92])

    print(f"Saved to {filename}")



  def move_camera(self, delta:np.ndarray):
    self.camera.move(delta)
    self.dirty = True

  def rotate_camera(self, delta:np.ndarray):
    self.camera.rotate(delta)
    self.dirty = True

  def set_camera_pose(self, r:np.ndarray, t:np.ndarray):
    self.camera.set_pose(r, t)
    self.dirty = True





    




from copy import deepcopy
from dataclasses import replace
from typing import Tuple
import trimesh
import pyrender

import numpy as np
from scipy.spatial.transform import Rotation as R

from splat_viewer.camera.fov import FOVCamera, join_rt


def normalize(v):
  return v / np.linalg.norm(v)


def to_pyrender_camera(camera:FOVCamera):
    fx, fy = camera.focal_length
    cx, cy = camera.principal_point

    pr_camera = pyrender.IntrinsicsCamera(fx, fy, cx, cy,
        znear=camera.near, zfar=camera.far)
    
    rotation = camera.rotation  @ flip_yz 
    m = join_rt(rotation, camera.position)

    return pyrender.Node(camera=pr_camera, matrix=m)


def look_at(eye, target, up=np.array([0., 0., 1.])):
  forward = normalize(target - eye)
  left = normalize(np.cross(up, forward))
  true_up = np.cross(forward, left)
  return np.stack([left, true_up, forward])


def look_at_pose(eye, target, up=np.array([0., 0., 1.])):
  pose = np.eye(4)
  pose[:3, :3] = look_at(eye, target, up)
  pose[:3, 3] = eye
  return pose

def make_sphere(pos, color, radius):
  sphere = trimesh.creation.icosphere(radius=radius)
  sphere.visual.vertex_colors = color
  sphere_mesh = pyrender.Mesh.from_trimesh(sphere)
  node = pyrender.Node(mesh=sphere_mesh, translation=pos)
  return node

def fov_to_focal(fov, image_size):
  return image_size / (2 * np.tan(fov / 2))

flip_yz = np.array([
  [1, 0, 0],
  [0, -1, 0],
  [0, 0, -1]
])


class SceneCamera:
  def __init__(self):

    self._camera = FOVCamera(position=np.array([0, 0, 0]),
                            rotation=np.eye(3),
                            focal_length=fov_to_focal(60, (640, 480)),
                            image_size = np.array((640, 480)),
                            image_name="viewport")

   
  def set_camera(self, camera):
    self._camera = deepcopy(camera)
   
 
  def look_at(self, pos:np.array, target:np.ndarray, up:np.ndarray=np.array([0, 0, 1])):
    self._camera = self._camera.with_pose(look_at_pose(pos, target, up))


  def resized(self, size:Tuple[int, int]):
    scale_factor = size[1] / self._camera.image_size[1]
    return self._camera.scale_size(scale_factor).pad_to(np.array(size))
  
  @property
  def view_matrix(self):
    return self._camera.world_t_camera

  @property 
  def rotation(self):
    return self._camera.rotation
  
  @rotation.setter
  def rotation(self, value):
    return self.set_pose(value, self.pos)

  @property 
  def pos(self):
    return self._camera.position
  
  @pos.setter
  def pos(self, value):
    return self.set_pose(self.rotation, value)

  
  def set_pose(self, r, t):
    self._camera = replace(self._camera, rotation=r, position=t)


  def move(self, delta:np.ndarray):
    self.pos += self.rotation @ delta

  def rotate(self, ypr):
    m = self.view_matrix
    m[:3, :3] =  m[:3, :3] @ R.from_euler('yxz', ypr).as_matrix()
     
    self.rotation = m[:3, :3]


  def zoom(self, factor):
    self._camera = self._camera.zoom(factor)



from typing import Optional, Set, Tuple
from PySide6 import QtGui
from PySide6.QtCore import QEvent

from beartype import beartype
import numpy as np
import torch
from splat_viewer.gaussians.data_types import Rendering

from splat_viewer.viewer.settings import Settings


class Interaction():
  def __init__(self):
    super(Interaction, self).__init__()
    self._child = None 
    self.active = False

  def transition(self, interaction:Optional['Interaction']):
    self.pop()
    if interaction is not None:
      self.push(interaction)

  def push(self, interaction:'Interaction'):
    self._child = interaction
    if self.active:
      self._child._activate()


  def pop(self):    
    if self._child is not None:
      child = self._child
      self._child = None
      child._deactivate()

  def transition(self, interaction:Optional['Interaction']):
    self.pop()
    if interaction is not None:
      self.push(interaction)

  def _activate(self):
    self.on_activate()

    if self._child is not None:
      self._child._activate()

    self.active = True


  def _deactivate(self):
    if self._child is not None:
      child = self._child
      self._child = None
      child._deactivate()

    self.on_deactivate()


  def trigger_event(self, event: QEvent) -> bool:
    if self._child is not None:
      if self._child.trigger_event(event):
        return True
      
    return self.event(event) 
  
  def _update(self, dt:float) -> bool:
    if self._child is not None:
      if self._child._update(dt):
        return True
    
    return self.update(dt)

  @beartype
  def event(self, event: QEvent) -> bool:
    event_callbacks = {
      QEvent.KeyPress: self.keyPressEvent,
      QEvent.KeyRelease: self.keyReleaseEvent,
      QEvent.MouseButtonPress: self.mousePressEvent,
      QEvent.MouseButtonRelease: self.mouseReleaseEvent,
      QEvent.MouseMove: self.mouseMoveEvent,
      QEvent.Wheel: self.wheelEvent,
      QEvent.FocusIn: self.focusInEvent,
      QEvent.FocusOut: self.focusOutEvent,
    }

    if event.type() in event_callbacks:
      return event_callbacks[event.type()](event) or False
    
    return False
    
  def trigger_paint(self, event: QtGui.QPaintEvent, view_changed:bool) -> bool:
    if self._child is not None:
      if self._child.trigger_paint(event, view_changed):
        return True
      
    return self.paintEvent(event, view_changed) 

  def keyPressEvent(self, event: QtGui.QKeyEvent):
    return False

  def keyReleaseEvent(self, event: QtGui.QKeyEvent):
    return False

  def mousePressEvent(self, event: QtGui.QMouseEvent):
    return False

  def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
    return False

  def mouseMoveEvent(self, event: QtGui.QMouseEvent):
    return False

  def wheelEvent(self, event: QtGui.QWheelEvent):
    return False

  def focusInEvent(self, event: QtGui.QFocusEvent):
    return False
  
  def focusOutEvent(self, event: QtGui.QFocusEvent):
    return False
  
  def paintEvent(self, event: QtGui.QPaintEvent, view_changed:bool):
    return False

  @beartype
  def update(self, dt) -> bool:
    return False

  def on_activate(self):
    pass

  def on_deactivate(self):
    pass

  @property
  def scene_widget(self):
    from .scene_widget import SceneWidget
    return SceneWidget.instance
  
  
  
  @property
  def settings(self) -> Settings: 
    return self.scene_widget.settings
  
  @property
  def modifiers(self) -> QtGui.Qt.KeyboardModifier:
    return self.scene_widget.modifiers
  
  @property
  def keys_down(self) -> Set[QtGui.Qt.Key]:
    return self.scene_widget.keys_down
  
  @property
  def cursor_pos(self) -> Tuple[int, int]:
    return self.scene_widget.cursor_pos
  
  @property
  def current_point(self) -> np.ndarray:
    return self.scene_widget.current_point_3d

  def lookup_point_3d(self, p:Tuple[int, int]) -> np.ndarray:
    return self.scene_widget.lookup_point_3d(p)

  def lookup_depth(self, p:Tuple[int, int]) -> np.ndarray:
    return self.scene_widget.lookup_depth(p)
  

  def lookup_depths(self, p:np.ndarray) -> np.ndarray:
    return self.scene_widget.lookup_depths(p)
  
  def test_depths(self, p:np.ndarray, depth:np.ndarray) -> np.ndarray:
    return self.scene_widget.test_depths(p, depth)
  
  def unproject_point(self, p:Tuple[int, int], depth:float) -> np.ndarray:
    return self.scene_widget.unproject_point(p, depth)
  
  def unproject_radius(self, p:Tuple[int, int], depth:float, radius:float
                        ) -> Tuple[np.ndarray, float]:
      return self.scene_widget.unproject_radius(p, depth, radius)
  
  @property
  def depth_map(self):
    return self.scene_widget.depth_map

  def from_numpy(self, a:np.ndarray):
    return torch.from_numpy(a).to(device=self.settings.device)

  @property
  def rendering(self) -> Rendering:
    return self.scene_widget.renderer.rendering
  

  @property
  def renderer(self):
    return self.scene_widget.renderer
  
  @property
  def gaussians(self):
    return self.scene_widget.renderer.gaussians

  def update_setting(self, **kwargs):
    self.scene_widget.update_setting(**kwargs)






from typing import List
from beartype import beartype
import numpy as np

from tqdm import tqdm

import torch
from .fov import FOVCamera

def make_homog(points):
  shape = list(points.shape)
  shape[-1] = 1
  return torch.concatenate([points, torch.ones(shape, dtype=torch.float32, device=points.device)], axis=-1)

def _transform_points(transform, points):
  assert points.shape[
      -1] == 3, 'transform_points: expected 3d points of ...x3, got:' + str(
          points.shape)

  homog = make_homog(points).reshape([-1, 4, 1])
  transformed = transform.reshape([1, 4, 4]) @ homog
  return transformed[..., 0].reshape(-1, 4)

def project_points(transform, xyz):
  homog = _transform_points(transform, xyz)
  depth = homog[..., 2:3]
  xy = homog[..., 0:2] 
  return (xy / depth), depth



@beartype
def visibility(cameras:List[FOVCamera], points:torch.Tensor, near=0.1, far=torch.inf):
  near_counts = torch.zeros(len(points), dtype=torch.int32, device=points.device)
  vis_counts = torch.zeros(len(points), dtype=torch.int32, device=points.device)

  projections = np.array([camera.projection for camera in cameras])
  torch_projections = torch.from_numpy(projections).to(dtype=torch.float32, device=points.device)

  for camera, proj in tqdm(zip(cameras, torch_projections)):
  
    proj, depth = project_points(proj, points)
    width, height = camera.image_size

    is_valid = ((proj[:, 0] >= 0) & (proj[:, 0] < width) & 
             (proj[:, 1] >= 0) & (proj[:, 1] < height)
             & (depth[:, 0] > near)
             )

    q = torch.quantile(depth[is_valid], 0.25)
    is_near = depth[:, 0] <= (far if np.isfinite(far) else 2 * q)

    near_counts[is_valid & is_near] += 1
    vis_counts[is_valid] += 1

  return near_counts, vis_counts





from functools import cached_property
from pathlib import Path
import re
from beartype import beartype
from natsort import natsorted
from dataclasses import dataclass

import numpy as np

from splat_viewer.camera.fov import FOVCamera, load_camera_json

from .loading import read_gaussians
from splat_viewer.gaussians import Gaussians

import open3d as o3d 

@beartype
@dataclass
class Workspace:
  model_path:Path

  cloud_files : dict[str, Path]
  cameras:list[FOVCamera]

  @cached_property
  def camera_extent(self):
    camera_positions = np.array([c.position for c in self.cameras])  

    scene_diagonal = np.linalg.norm(
      camera_positions.max(axis=0) - camera_positions.min(axis=0))

    return scene_diagonal / 2.0


  def latest_iteration(self) -> str:
    paths = [(m.group(1), name) for name in self.cloud_files.keys()
             if (m:=re.search("iteration_(\d+)", name))]
    
    if len(paths) == 0:
      raise FileNotFoundError(f"No point clouds named iteration_(\d+) in {str(self.model_path)}")

    paths = sorted(paths, key=lambda x: int(x[0]))
    return paths[-1][1]

  def model_filename(self, model:str) -> Path:
    if model not in self.cloud_files:
      options = list(self.cloud_files.keys())
      raise LookupError(f"Model {model} not found in {self.model_path} options are: {options}")

    return self.cloud_files[model]      

  def load_model(self, model:str) -> Gaussians:
    return read_gaussians(self.model_filename(model))
  
  def load_initial_points(self) -> o3d.t.geometry.PointCloud:
    return o3d.t.io.read_point_cloud(str(self.model_path / "input.ply"))
  

def find_clouds(p:Path):
  clouds = {d.name : file for d in p.iterdir() 
            if d.is_dir() and (file :=d / "point_cloud.ply").exists()
  }

  if len(clouds) == 0:
    raise FileNotFoundError(f"No point clouds found in {str(p)}")

  return clouds


def load_workspace(model_path:Path | str) -> Workspace:
  model_path = Path(model_path)
  cloud_path = model_path / "point_cloud"
  
  if not cloud_path.exists():
    raise FileNotFoundError(f"Could not find point cloud directory at {str(cloud_path)}")
  
  cloud_files = find_clouds(cloud_path)

  cameras = load_camera_json(model_path / "cameras.json")
  cameras = natsorted(cameras.values(), key=lambda x: x.image_name)

  return Workspace(
    model_path = model_path,
    cloud_files = cloud_files,
    cameras = cameras
  )




import numpy as np


def expand_identity(m, shape=(4, 4)):
  expanded = np.eye(*shape)
  expanded[0:m.shape[0], 0:m.shape[1]] = m
  return expanded


def join_rt(r, t):
  assert t.ndim == r.ndim - 1 and t.shape[-1] == 3 and r.shape[-2:] == (3, 3),\
    'got r.shape:' + str(r.shape) + ' and t.shape:' + str(t.shape)

  d = t.ndim
  m_34 = np.concatenate([r, np.expand_dims(t, d)], axis=d)
  row = np.broadcast_to(np.array([[0, 0, 0, 1]]), (*r.shape[:-2], 1, 4))
  return np.concatenate([m_34, row], axis=d - 1)


def split_rt(m):
  assert m.shape[-2:] == (4, 4)
  return m[..., :3, :3], m[..., :3, 3]


def translation(m):
  assert m.shape[-2:] == (4, 4)
  return m[..., :3, 3]


def rotation(m):
  assert m.shape[-2:] == (4, 4)
  return m[..., :3, :3]


def cross_matrix(vec):
  return np.cross(vec, np.identity(vec.shape[0]) * -1)


def make_grid(width, height):
  return np.meshgrid(np.arange(0, width, dtype=np.float32),
                     np.arange(0, height, dtype=np.float32))


def make_grid_centred(width, height):
  return np.meshgrid(np.arange(0.5, width, dtype=np.float32),
                     np.arange(0.5, height, dtype=np.float32))


def transform_grid_homog(x, y, z, w, transform):
  points = np.stack([x, y, z, w], axis=2)
  transformed = points.reshape(
      (-1, 4)) @ np.transpose(transform.astype(np.float32))
  return transformed.reshape(*z.shape, 4)


def make_homog(points):
  shape = list(points.shape)
  shape[-1] = 1
  return np.concatenate([points, np.ones(shape, dtype=np.float32)], axis=-1)


def transform_grid(x, y, z, transform):
  """  transform points of (x, y, z) by 4x4 matrix """
  return transform_grid_homog(x, y, z, np.ones(z.shape, dtype=np.float32),
                              transform)


def uproject_invdepth(invdepth: np.ndarray, depth_t_disparity: np.ndarray):
  """ perspective transform points of (x, y, 1/depth) by 4x4 matrix """
  x, y = make_grid(invdepth.shape[1], invdepth.shape[0])
  return transform_grid_homog(x, y, np.ones(x.shape, dtype=x.dtype), invdepth,
                              depth_t_disparity)


def uproject_depth(depth, transform):
  x, y = make_grid(depth.shape[1], depth.shape[0])
  return transform_grid_homog(x * depth, y * depth, depth,
                              np.ones(x.shape, dtype=x.dtype), transform)


def uproject_disparity(disparity, transform):
  x, y = make_grid(disparity.shape[1], disparity.shape[0])
  return transform_grid_homog(x, y, disparity, np.ones(x.shape, dtype=x.dtype),
                              transform)


def transform_invdepth(invdepth, depth_t_disparity):
  """ transform image grid of inverse-depth image by 4x4 matrix
    returns: inverse-depth image in new coordinate system 
    """
  points = uproject_invdepth(invdepth, depth_t_disparity)
  return points[:, :, 3] / points[:, :, 2]


def transform_depth(depth, depth_t_disparity):
  """ transform depth image by 4x4 matrix
    returns: image in new coordinate system 
    """
  points = uproject_invdepth(1 / depth, depth_t_disparity)
  return points[:, :, 2] / points[:, :, 3]


def _batch_transform(transforms, points):
  assert points.shape[
      -1] == 3 and points.ndim == 2, 'transform_points: expected 3d points of Nx3, got:' + str(
          points.shape)
  assert transforms.shape[-2:] == (
      4, 4
  ) and transforms.ndim == 3, 'transform_points: expected Mx4x4, got:' + str(
      transforms.shape)

  homog = make_homog(points)
  transformed = transforms.reshape(transforms.shape[0], 1, 4, 4) @ homog.reshape(
      1, *homog.shape, 1)

  return transformed[..., 0].reshape([transforms.shape[0], -1, 4])


def batch_transform_points(transforms, points):
  return _batch_transform(transforms, points)[..., 0:3]


def batch_project_points(transforms, points):
  homog = _batch_transform(transforms, points) 
  return homog[..., 0:3] / homog[..., 3:4]


def _transform_points(transform, points):
  assert points.shape[
      -1] == 3, 'transform_points: expected 3d points of ...x3, got:' + str(
          points.shape)

  homog = make_homog(points).reshape([-1, 4, 1])
  transformed = transform.reshape([1, 4, 4]) @ homog
  return transformed[..., 0].reshape(-1, 4)


def transform_points(transform, points):
  return _transform_points(transform, points)[..., 0:3]

def unproject_pixels(transform, xy, depth):
  points = np.concatenate([xy * depth, depth], axis=-1)

  homog = _transform_points(transform, points)
  return  homog[..., 0:3] / homog[..., 3:4]

def project_points(transform, xyz):
  homog = _transform_points(transform, xyz)
  depth = homog[..., 2:3]
  xy = homog[..., 0:2] 
  return (xy / depth), depth

def affine_transform_points(transform, points):
  assert points.shape[
      -1] == 3, 'affine_transform_points: expected 3d points of ...x3, got:' + str(
          points.shape)

  r, t = split_rt(transform)
  transformed = (r.reshape(1, 3, 3) @ points.reshape(-1, 3, 1)).reshape(-1,
                                                                        3) + t

  return transformed.reshape(points.shape)


def translate_33(tx, ty):
  return np.array([[1, 0, tx], [0, 1, ty], [0, 0, 1]])


def scale_33(sx, sy):
  return np.diag([sx, sy, 1])


def translate_44(tx, ty, tz):
  return np.array([
      [1, 0, 0, tx],
      [0, 1, 0, ty],
      [0, 0, 1, tz],
      [0, 0, 0, 1],
  ])


def scale_44(sx, sy, sz):
  return np.diag([sx, sy, sz, 1])


def dict_subset(dict, keys):
  return {k: dict[k] for k in keys}


def check_size(image, expected_size):
  size = image.shape[1], image.shape[0]
  assert size == tuple(
      expected_size), 'got: ' + str(size) + ', should be: ' + str(expected_size)


def estimate_rigid_transform(ref_points, points):
  assert ref_points.shape == points.shape and points.shape[0] >= 3,\
     'estimate_transform: expected at least 3 points, got:' + str(points.shape)

  centroid_ref = np.mean(ref_points, axis=0)
  centroid_points = np.mean(points, axis=0)

  centered_ref = ref_points - centroid_ref
  centered_points = points - centroid_points

  s = centered_ref.T @ centered_points

  u, s, vh = np.linalg.svd(s)
  r = np.dot(vh.T, u.T)

  if np.linalg.det(r) < 0:
    vh[-1, :] *= -1
    r = np.dot(vh.T, u.T)

  t = centroid_points - np.dot(r, centroid_ref)
  transform = join_rt(r, t)

  err = np.linalg.norm(transform_points(transform, ref_points) - points, axis=1)
  return transform, err



def normalize(v):
  return v / np.linalg.norm(v, axis=-1)




import argparse
import itertools
from pathlib import Path
import time

import torch
from tqdm import tqdm


from splat_viewer.gaussians.data_types import Gaussians
from splat_viewer.gaussians.workspace import load_workspace
from splat_viewer.renderer.arguments import add_render_arguments, renderer_from_args

import taichi as ti
from torch.profiler import profile, record_function, ProfilerActivity



def bench_forward(gaussians, renderer, cameras, profiler=None, **kwargs):
    inputs = renderer.pack_inputs(gaussians)
    torch.cuda.synchronize()

    start_time = time.time()
    with torch.no_grad():
      for camera in tqdm(cameras):
        renderer.render(inputs, camera, **kwargs)

        if profiler is not None:
          profiler.step()

    torch.cuda.synchronize()
    return time.time() - start_time

def bench_backward(gaussians, renderer, cameras, profiler=None, **kwargs):
    inputs = renderer.pack_inputs(gaussians, requires_grad=True)
    torch.cuda.synchronize()

    start_time = time.time()
    for camera in tqdm(cameras):

      output = renderer.render(inputs, camera, **kwargs)
      loss = output.image.sum()
      loss.backward()

      if profiler is not None:
        profiler.step()

    torch.cuda.synchronize()
    return time.time() - start_time

def main():

  torch.set_printoptions(precision=5, sci_mode=False, linewidth=120)

  parser = argparse.ArgumentParser()

  parser.add_argument("model_path", type=Path, help="workspace folder containing cameras.json, input.ply and point_cloud folder with .ply models")
  parser.add_argument("--device", type=str, default="cuda:0", help="torch device to use")
  parser.add_argument("--model", type=str, default=None, help="model iteration to load from point_clouds folder")

  parser.add_argument("--profile", action="store_true", help="enable profiling")
  parser.add_argument("--trace", type=str, default=None, help="enable profiling with tensorboard trace for profiled data")
  parser.add_argument("--backward", action="store_true", help="benchmark backward pass")

  add_render_arguments(parser)

  parser.add_argument("--debug", action="store_true", help="enable taichi kernels in debug mode")
  parser.add_argument("-n", type=int, default=500, help="number of iterations to render")

  parser.add_argument("--sh_degree", type=int, default=None, help="modify spherical harmonics degree")
  parser.add_argument("--no_sort", action="store_true", help="disable sorting by scale (sorting makes tilemapping faster)")



  parser.add_argument("--depth", action="store_true", help="render depth maps")
  parser.add_argument("--image_size", type=int, default=None, help="resize longest edge of camera image sizes")


  
  args = parser.parse_args()

  ti.init(arch=ti.cuda, offline_cache=True, log_level=ti.INFO,
          debug=args.debug, device_memory_GB=0.1)


  workspace = load_workspace(args.model_path)
  if args.model is None:
      args.model = workspace.latest_iteration()
    
  gaussians:Gaussians = workspace.load_model(args.model).to(args.device)
  if args.sh_degree is not None:
    gaussians = gaussians.with_sh_degree(args.sh_degree)

  # sorting by scale makes the tile mapping algorithm more efficient
  if not args.no_sort:
    gaussians = gaussians.sorted()

  renderer = renderer_from_args(args)

  cameras = workspace.cameras
  if args.image_size is not None:
    cameras = [camera.resize_longest(args.image_size) for camera in cameras]

  def n_cameras(n):
    return list(itertools.islice(itertools.cycle(cameras), n))

  print(f"Benchmarking {args.model_path} with {gaussians.batch_size[0]} points")

  image_sizes = set([tuple(camera.image_size) for camera in cameras])
  print(f"Cameras: {len(workspace.cameras)}, Image sizes: {image_sizes}")


  bench_renders = bench_backward if args.backward else bench_forward

  warmup_size = args.n // 10
  print(f"Warmup @ {warmup_size} cameras")
  bench_renders(gaussians, renderer, n_cameras(warmup_size), render_depth=args.depth)


  print(f"Benchmark @ {args.n} cameras:")


  if args.profile or args.trace:
    with profile(activities=[ProfilerActivity.CUDA], 
                record_shapes=True, with_stack=True) as prof:
      with record_function("model_inference"):
        ellapsed = bench_renders(gaussians, renderer, n_cameras(args.n), profiler=prof, render_depth=args.depth)
    result = prof.key_averages(group_by_stack_n=4).table(sort_by="self_cuda_time_total", 
                                       row_limit=25, max_name_column_width=100)
    
    if args.trace:
      print(f"Writing chrome trace file to {args.trace}")
      prof.export_chrome_trace(args.trace)

    if args.profile:
      print(result)

  else:
    ellapsed = bench_renders(gaussians, renderer, n_cameras(args.n), render_depth=args.depth)
  print(f"{args.n} images in {ellapsed:.2f}s at {args.n / ellapsed:.2f} images/s")

    

if __name__ == "__main__":
  main()  





from dataclasses import replace
import math
from beartype import beartype
import torch


from taichi_3d_gaussian_splatting.GaussianPointCloudRasterisation \
  import GaussianPointCloudRasterisation, CameraInfo 

from taichi_3d_gaussian_splatting.utils import SE3_to_quaternion_and_translation_torch

import torch.nn.functional as F
import numpy as np

from splat_viewer.camera.fov import FOVCamera
from splat_viewer.gaussians.data_types import Gaussians, Rendering
from splat_viewer.viewer.scene_camera import fov_to_focal


def pad_to_tile(camera:FOVCamera, m:int):
  def next_mult(x):
      return int(math.ceil(x / m) * m)
      
  w, h = camera.image_size

  round_size = (next_mult(w), next_mult(h))
  return camera.pad_to(np.array(round_size))


default_camera = FOVCamera(position=np.array([0, 0, 0]),
                          rotation=np.eye(3),
                          focal_length=fov_to_focal(60, (640, 480)),
                          image_size = np.array((640, 480)),
                          image_name="viewport")

class TaichiRenderer:
  def __init__(self):
    config = GaussianPointCloudRasterisation.GaussianPointCloudRasterisationConfig(
      near_plane=0.1,
      far_plane=1000.0,
      depth_to_sort_key_scale=10000.0,
    )

    self.renderer = GaussianPointCloudRasterisation(config)

  def camera_info(self, camera:FOVCamera, device=torch.device("cuda:0")):
    return CameraInfo(
      torch.from_numpy(camera.intrinsic).to(dtype=torch.float32, device=device),
      camera_width=camera.image_size[0],
      camera_height=camera.image_size[1],
      camera_id = 0
    )
  
  def camera_qt(self, camera:FOVCamera, device):
    return SE3_to_quaternion_and_translation_torch(
          torch.from_numpy(camera.world_t_camera
          ).unsqueeze(0).to(dtype=torch.float32, device=device))

  @beartype
  def with_camera(self, camera:FOVCamera, inputs:GaussianPointCloudRasterisation.GaussianPointCloudRasterisationInput):
    q, t = self.camera_qt(camera, inputs.point_cloud.device)
    return replace(inputs,
      camera_info = self.camera_info(camera, device=inputs.point_cloud.device),
      q_pointcloud_camera=q.contiguous(),
      t_pointcloud_camera=t.contiguous()
    )
      
  @beartype 
  def as_features(self, gaussians:Gaussians):
    n_points, n_feats, _ = gaussians.sh_feature.shape
    sh_features = F.pad(gaussians.sh_feature.permute(0, 2, 1), (0, 16 - n_feats))

    return torch.concatenate([
        gaussians.rotation,
        gaussians.log_scaling,
        gaussians.alpha_logit,
        sh_features.reshape(n_points, -1)
        ], dim=1)

  @beartype
  def pack_inputs(self, gaussians:Gaussians, requires_grad=False):

    return GaussianPointCloudRasterisation.GaussianPointCloudRasterisationInput(
      point_cloud = gaussians.position.requires_grad_(requires_grad),
      point_cloud_features = self.as_features(gaussians).requires_grad_(requires_grad),

      camera_info = self.camera_info(default_camera, device=gaussians.device),
      q_pointcloud_camera=torch.eye(4, dtype=torch.float32, device=gaussians.device),
      t_pointcloud_camera=torch.zeros(3, dtype=torch.float32, device=gaussians.device),

      point_object_id = torch.zeros(gaussians.position.shape[0], dtype=torch.int32),
      point_invalid_mask = torch.zeros(gaussians.position.shape[0], dtype=torch.int8), 
      color_max_sh_band = gaussians.sh_degree(),
    )
    

  def render(self, inputs:GaussianPointCloudRasterisation.GaussianPointCloudRasterisationInput, 
             camera:FOVCamera, render_depth:bool=False):
    self.renderer.config = replace(self.renderer.config, 
        near_plane=camera.near, far_plane=camera.far,
        rgb_only=not render_depth
        )
    
    w, h = camera.image_size
    padded = pad_to_tile(camera, 16)
    inputs = self.with_camera(padded, inputs)

    image, depth, _ = self.renderer.forward(inputs)
    return Rendering(image=image[:h, :w], 
                     depth=depth[:h, :w] if depth is not None else None, 
                     camera=camera)



import argparse
from dataclasses import replace
from pathlib import Path
from typing import List

import torch
from splat_viewer.camera.fov import FOVCamera
from splat_viewer.camera.visibility import visibility

from splat_viewer.gaussians.loading import  read_gaussians, to_pcd, write_gaussians
from splat_viewer.gaussians.workspace import load_workspace


def crop_model(model, cameras:List[FOVCamera], args):
  is_near, is_visible = visibility(cameras, model.positions, near = args.near, far = args.far)

  min_views = max(1, len(cameras) * args.min_percent / 100)
  n = (is_visible > 0).sum(dtype=torch.int32)
  n_near = (is_near >= min_views).sum(dtype=torch.int32)

  print(f"Cropped model from {model.batch_shape} to {n} visible points, {n_near} near (at least {min_views} views)")
  model = replace(model, foreground=(is_near >= min_views).reshape(-1, 1))

  model = model[is_visible > 0]
  return model

def main():

  parser = argparse.ArgumentParser(description="Add a 'foreground' annotation to a .ply gaussian splatting file")
  parser.add_argument("model_path", type=Path, help="Path to the gaussian splatting workspace")
  parser.add_argument("--scan", type=str,  help="Input scan file")
  
  parser.add_argument("--far", default=torch.inf, type=float, help="Max depth to determine the visible ROI")
  parser.add_argument("--near", default=0.2, type=float, help="Min depth to determine the visible ROI")


  parser.add_argument("--min_percent", type=float, default=0, help="Minimum percent of views to be included")
  parser.add_argument("--device", default='cuda:0')

  parser.add_argument("--statistical_outliers", type=float, default=None)
  parser.add_argument("--radius_outliers", type=float, default=None)
  parser.add_argument("--knn", type=int, default=100)

  
  parser.add_argument("--write", action="store_true", help="Write the cropped model to a file")
  parser.add_argument("--show", action="store_true")

  args = parser.parse_args()

  assert args.show or args.write, "Nothing to do. Please specify --show or --write"


  workspace = load_workspace(args.model_path)

  with torch.inference_mode():
    workspace = load_workspace(args.model_path)

    model_name = workspace.latest_iteration()
    model_file = workspace.model_filename(model_name)
    model = read_gaussians(model_file)

  
    model = model.to(args.device)
    model = crop_model(model, workspace.cameras, args)
    
    idx_crop = torch.nonzero(model.foreground.squeeze() > 0).squeeze()

    def filtered(mask):
      mask = torch.from_numpy(mask.numpy()).to(model.device)
      model.foreground[idx_crop[~mask], :] = False

    if any([args.radius_outliers, args.statistical_outliers]):
      pcd = to_pcd(model[idx_crop].cpu())

      if args.statistical_outliers is not None:
        _, keep = pcd.remove_statistical_outliers(nb_neighbors=args.knn, std_ratio=args.statistical_outliers)
        filtered(keep)

      if args.radius_outliers is not None:
        _, keep = pcd.remove_radius_outliers(nb_points=args.knn, search_radius=args.radius_outliers)
        filtered(keep)


      num_removed = idx_crop.shape[0] - (model.foreground > 0).sum()
      print(f"Found {num_removed} outliers")

    if args.write:    
      write_gaussians(model_file, model)
      print(f"Wrote {model} to {model_file}")

    if args.show:
      from splat_viewer.viewer.viewer import show_workspace
      show_workspace(workspace, model)



if __name__ == "__main__":
  main()  









import math
from typing import List
import trimesh

import numpy as np
import pyrender

from splat_viewer.camera.fov import FOVCamera
from splat_viewer.camera.transforms import batch_transform_points

def instance_meshes(mesh:trimesh.Trimesh, transforms:np.array):
  vertices = batch_transform_points(transforms, mesh.vertices)
  n = transforms.shape[0]

  offsets = np.arange(n).reshape(n, 1, 1) * mesh.vertices.shape[0] 
  faces = mesh.faces.reshape(1, -1, 3) + offsets

  return trimesh.Trimesh(vertices=vertices.reshape(-1, 3), 
                         faces=faces.reshape(-1, 3))

def camera_marker(camera:FOVCamera, scale):
  fov = camera.fov

  x = math.tan(fov[0] / 2)
  y = math.tan(fov[1] / 2) 

  points = np.array([
    [0, 0, 0],
    [-x,  y, 1],
    [x,   y, 1],
    [x,  -y, 1],
    [-x, -y, 1]
    ])
  

  triangles = np.array([
    [0, 1, 2],
    [0, 2, 3],
    [0, 3, 4],
    [0, 4, 1],

    [1, 2, 3],
    [1, 3, 4]
  ], dtype=np.int32) 

  return trimesh.Trimesh(vertices=points * scale, faces=triangles, process=False) 
    

def make_camera_markers(cameras:List[FOVCamera], scale:float):
  mesh = camera_marker(cameras[0], scale)

  markers = instance_meshes(mesh, np.array([cam.world_t_camera for cam in cameras], dtype=np.float32))
  markers = pyrender.Mesh.from_trimesh(markers, wireframe=True, smooth=False,
                  material=pyrender.MetallicRoughnessMaterial
                  (doubleSided=True, wireframe=True, smooth=False, baseColorFactor=(255, 0, 0, 255)))


  return markers


def make_sphere(radius=1.0, subdivisions=3, color=(0.0, 0.0, 1.0)):
  sphere = trimesh.creation.icosphere(radius=radius, subdivisions=subdivisions)

  material = pyrender.MetallicRoughnessMaterial(
    metallicFactor=0.0,
    baseColorFactor=[*color, 1.0],
  )
  return pyrender.Mesh.from_trimesh(sphere, smooth=True, material=material)


import math

def check_sh_degree(sh_features):
  assert len(sh_features.shape) == 3, f"SH features must have 3 dimensions, got {sh_features.shape}"

  n_sh = sh_features.shape[2]
  n = int(math.sqrt(n_sh))

  assert n * n == n_sh, f"SH feature count must be square, got {n_sh} ({sh_features.shape})"
  return (n - 1)

def num_sh_features(deg):
  return (deg + 1) ** 2 


sh0 = 0.282094791773878

def rgb_to_sh(rgb):
    return (rgb - 0.5) / sh0

def sh_to_rgb(sh):
    return sh * sh0 + 0.5


from dataclasses import replace
import math
from typing import Optional
from tensordict import tensorclass
import torch

from .sh_utils import check_sh_degree, num_sh_features, rgb_to_sh, sh_to_rgb

from dataclasses import dataclass
from splat_viewer.camera.fov import FOVCamera



@dataclass
class Rendering:
  image : torch.Tensor
  camera : FOVCamera

  depth : Optional[torch.Tensor] = None
  depth_var: Optional[torch.Tensor] = None

  @property
  def image_size(self):
    y, x = self.image.shape[1:]
    return x, y

@tensorclass
class Gaussians():
  position     :  torch.Tensor # 3  - xyz
  log_scaling   : torch.Tensor # 3  - scale = exp(log_scalining) 
  rotation      : torch.Tensor # 4  - quaternion wxyz
  alpha_logit   : torch.Tensor # 1  - alpha = sigmoid(alpha_logit)
  
  sh_feature      :  torch.Tensor # (any rgb (3), spherical harmonics (3x16) etc)
  
  foreground : Optional[torch.Tensor] # 1 (bool) 
  label      : Optional[torch.Tensor] # 1 (int)

  def __post_init__(self):
    assert self.position.shape[1] == 3, f"Expected shape (N, 3), got {self.position.shape}"
    assert self.log_scaling.shape[1] == 3, f"Expected shape (N, 3), got {self.log_scaling.shape}"
    assert self.rotation.shape[1] == 4, f"Expected shape (N, 4), got {self.rotation.shape}"
    assert self.alpha_logit.shape[1] == 1, f"Expected shape (N, 1), got {self.alpha_logit.shape}"

    check_sh_degree(self.sh_feature)
    assert self.foreground is None or self.foreground.shape[1] == 1, f"Expected shape (N, 1), got {self.foreground.shape}"
    assert self.label is None or self.label.shape[1] == 1, f"Expected shape (N, 1), got {self.label.shape}"

  def n(self):
    return self.batch_size[0]

  def __repr__(self):
    return f"Gaussians with {self.batch_shape[0]} points, sh_degree={self.sh_degree}"

  def __str__(self):
    return repr(self)
    
  def packed(self):
    return torch.cat([self.position, self.log_scaling, self.rotation, self.alpha_logit], dim=-1)

  def scale(self):
    return torch.exp(self.log_scaling)

  def alpha(self):
    return torch.sigmoid(self.alpha_logit)
  
  def sh_degree(self):
    return check_sh_degree(self.sh_feature)
  
  def with_fixed_scale(self, scale:float):
    return replace(self, 
            log_scaling=torch.full_like(self.log_scaling, math.log(scale)),
            batch_size=self.batch_size)
  
  def get_colors(self):
    return sh_to_rgb(self.sh_feature[:, :, 0])
  

  def with_colors(self, colors):
    sh_feature = self.sh_feature.clone()
    sh_feature[:, :, 0] = rgb_to_sh(colors)

    return replace(self, 
            sh_feature=sh_feature,
            batch_size=self.batch_size)
  
  def with_sh_degree(self, sh_degree:int):
    assert sh_degree >= 0

    if sh_degree <= self.sh_degree():
      return replace(self, 
          sh_feature = self.sh_feature[:, :, :num_sh_features(sh_degree)],
          batch_size=self.batch_size)
    else:
      num_extra = num_sh_features(sh_degree) - num_sh_features(self.sh_degree)
      extra_features = torch.zeros((self.batch_shape[0], 
              3, num_extra), device=self.device)
      
      return replace(self, sh_feature = torch.cat(
        [self.sh_feature, extra_features], dim=2), 
        batch_size=self.batch_size)
    
  def sorted(self):
    max_axis = torch.max(self.log_scaling, dim=1).values
    indices = torch.argsort(max_axis, descending=False)

    return self[indices]
    





from pathlib import Path
import tempfile
import open3d as o3d
import open3d.core as o3c

import numpy as np
import torch
import torch.nn.functional as F

from .data_types import Gaussians


def torch_to_o3d(tensor:torch.Tensor) -> o3d.core.Tensor:
  return o3d.core.Tensor.from_dlpack(torch.utils.dlpack.to_dlpack(tensor))

def o3d_to_torch(tensor:o3c.Tensor) -> torch.Tensor:
  return torch.from_dlpack(o3d.core.Tensor.to_dlpack(tensor))



def to_pcd(gaussians:Gaussians) -> o3d.t.geometry.PointCloud:
  pos = torch_to_o3d(gaussians.positions)

  pcd = o3d.t.geometry.PointCloud(pos.device)
  pcd.point['positions'] = pos
  pcd.point['opacity'] = torch_to_o3d(gaussians.alpha_logit)

  sh_dc, sh_rest = gaussians.split_sh()

  sh_dc = sh_dc.view(-1, 3)
  sh_rest = sh_rest.permute(0, 2, 1).reshape(sh_rest.shape[0], sh_rest.shape[1] * sh_rest.shape[2])

  for i in range(3):
    pcd.point[f'f_dc_{i}'] = torch_to_o3d(sh_dc[:, i:i+1])

  for i in range(sh_rest.shape[-1]):
    pcd.point[f'f_rest_{i}'] = torch_to_o3d(sh_rest[:, :, i:i+1])
  

  for i in range(3):
    pcd.point[f'scale_{i}'] = torch_to_o3d(gaussians.log_scaling[:, i:i+1])


  # convert back to wxyz quaternion
  rotation = torch.roll(gaussians.rotation, 1, dims=(1,))

  for i in range(4):
    pcd.point[f'rot_{i}'] = torch_to_o3d(rotation[:, i:i+1])

  if gaussians.foreground is not None:
    pcd.point['foreground'] = torch_to_o3d(gaussians.foreground.to(torch.int32))

  if gaussians.label is not None:
    pcd.point['label'] = torch_to_o3d(gaussians.label)
  

  return pcd



def from_pcd(pcd:o3d.t.geometry.PointCloud) -> Gaussians:
  def get_keys(ks):
    values = [o3d_to_torch(pcd.point[k]) for k in ks]
    return torch.concat(values, dim=-1)

  positions = o3d_to_torch(pcd.point['positions'])

  attrs = sorted(dir(pcd.point))
  sh_attrs = [k for k in attrs if k.startswith('f_rest_') or k.startswith('f_dc_')]
  
  n_sh = len(sh_attrs) // 3
  deg = int(np.sqrt(n_sh))

  assert deg * deg == n_sh, f"SH feature count must be (3x) square, got {len(sh_attrs)}"
  log_scaling = get_keys([f'scale_{k}' for k in range(3)])


  sh_dc = get_keys([f'f_dc_{k}' for k in range(3)]).view(positions.shape[0], 3, 1)
  sh_rest = get_keys([f'f_rest_{k}' for k in range(3 * (deg * deg - 1))])
  sh_rest = sh_rest.view(positions.shape[0], 3, n_sh - 1)

  sh_features = torch.cat([sh_dc, sh_rest], dim=2)  

  rotation = get_keys([f'rot_{k}' for k in range(4)])
  # convert from wxyz to xyzw quaternion and normalize
  rotation = torch.roll(F.normalize(rotation), -1, dims=(1,))
  
  alpha_logit = get_keys(['opacity'])

  
  foreground = (get_keys(['foreground']).to(torch.bool) 
    if 'foreground' in pcd.point else None)
  
  label = get_keys(['label']) if 'label' in pcd.point else None
  

  return Gaussians(
    position = positions, 
    rotation = rotation,
    alpha_logit = alpha_logit,
    log_scaling = log_scaling,
    sh_feature = sh_features,

    foreground = foreground,
    label = label,

    batch_size = (positions.shape[0],)
  )


def write_gaussians(filename:Path | str, gaussians:Gaussians):
  filename = Path(filename)

  pcd = to_pcd(gaussians)
  o3d.t.io.write_point_cloud(str(filename), pcd)

def read_gaussians(filename:Path | str) -> Gaussians:
  filename = Path(filename) 

  pcd:o3d.t.geometry.PointCloud = o3d.t.io.read_point_cloud(str(filename))
  if 'positions' not in pcd.point:
    raise ValueError(f"Could not load point cloud from {filename}")

  return from_pcd(pcd)
  

def random_gaussians(n:int, sh_degree:int):
  points = torch.randn(n, 3)

  return Gaussians( 
    position = points,
    rotation = F.normalize(torch.randn(n, 4), dim=1),
    alpha_logit = torch.randn(n, 1),
    log_scaling = torch.randn(n, 3) * 4,

    sh_feature = torch.randn(n, (sh_degree + 1)**2, 3),
  )

def test_read_write():
  temp_path = Path(tempfile.mkdtemp())

  print("Testing write/read")
  for i in range(10):
    g = random_gaussians((i + 1) * 1000, 3)
    write_gaussians(temp_path / f'gaussians_{i}.ply', g)
    g2 = read_gaussians(temp_path / f'gaussians_{i}.ply')

    assert torch.allclose(g.position, g2.position)
    assert torch.allclose(g.rotation, g2.rotation)
    assert torch.allclose(g.alpha, g2.alpha)
    assert torch.allclose(g.log_scaling, g2.log_scaling)
    assert torch.allclose(g.sh_feature, g2.sh_feature)





if __name__ == '__main__':
  test_read_write()





from dataclasses import dataclass, replace
import cv2

import numpy as np
import pyrender
import torch
from splat_viewer.camera.fov import FOVCamera
#from splat_viewer.renderer.diff_gaussian_rasterization import DiffGaussianRenderer
from splat_viewer.renderer.taichi_splatting import GaussianRenderer

from splat_viewer.gaussians.workspace import Workspace
from splat_viewer.gaussians import Gaussians, Rendering
from splat_viewer.viewer.scene_camera import to_pyrender_camera

    
from .mesh import make_camera_markers
from .settings import Settings, ViewMode






def get_cv_colormap(cmap):
  colormap = np.arange(256, dtype=np.uint8).reshape(1, -1)
  colormap = cv2.applyColorMap(colormap, cmap)
  return colormap.astype(np.uint8)


class PyrenderScene:
  
  def __init__(self,  workspace:Workspace):
    self.initial = workspace.load_initial_points()

    self.initial_scene = pyrender.Scene()

    self.points = pyrender.Mesh.from_points(
      self.initial.point['positions'].numpy(), self.initial.point['colors'].numpy())
    self.initial_node = self.initial_scene.add(self.points, pose=np.eye(4))

    self.initial_scene.ambient_light = np.array([1.0, 1.0, 1.0, 1.0])

    self.cameras = make_camera_markers(workspace.cameras, workspace.camera_extent / 50.)
    self.initial_scene.add(self.cameras)
    
    self.renderer = None


  def create_renderer(self, camera, settings:Settings):
    if self.renderer is None:
      self.renderer = pyrender.OffscreenRenderer(camera.image_size[0], camera.image_size[1], point_size=settings.point_size)
    else:
      self.renderer.viewport_width = camera.image_size[0]
      self.renderer.viewport_height = camera.image_size[1]
      self.renderer.point_size = settings.point_size

    return self.renderer

  def render(self, camera, settings:Settings):
    
    renderer = self.create_renderer(camera, settings)

    self.cameras.is_visible = settings.show.cameras
    self.points.is_visible = settings.show.initial_points

    node = to_pyrender_camera(camera)
    scene =  self.initial_scene
    scene.add_node(node)

    image, depth = renderer.render(scene)
    scene.remove_node(node)

    return image, depth

@dataclass(frozen=True)
class RenderState:
  as_points : bool = False
  cropped : bool = False

  def update_setting(self, settings:Settings):
    return replace(self,
      as_points = settings.view_mode == ViewMode.Points,
      cropped = settings.show.cropped)


  def updated(self, gaussians:Gaussians) -> Gaussians:

    if self.as_points:
      gaussians = gaussians.with_fixed_scale(0.001)

    if self.cropped and gaussians.foreground is not None:
      gaussians = gaussians[gaussians.foreground.squeeze()]

    return gaussians

class WorkspaceRenderer:
  def __init__(self, workspace:Workspace, gaussians:Gaussians, gaussian_renderer):
    self.workspace = workspace

    self.gaussians = gaussians
    self.packed_gaussians = None
    self.render_state = RenderState()

    self.gaussian_renderer = gaussian_renderer

    self.pyrender_scene = PyrenderScene(workspace)

    self.rendering = None
    self.color_map = torch.from_numpy(get_cv_colormap(cv2.COLORMAP_TURBO)
                                      ).squeeze(0).to(device=self.gaussians.device)


  def render_gaussians(self, camera, settings:Settings) -> Rendering:
    render_state = self.render_state.update_setting(settings)

    if self.packed_gaussians is None or self.render_state != render_state:
      self.packed_gaussians = self.gaussian_renderer.pack_inputs(
        render_state.updated(self.gaussians))
      
      self.render_state = render_state
    return self.gaussian_renderer.render(self.packed_gaussians, camera)


  def unproject_mask(self, camera:FOVCamera, 
                mask:torch.Tensor, alpha_multiplier=1.0, threshold=1.0):
    return self.gaussian_renderer.unproject_mask(self.gaussians, 
          camera, mask, alpha_multiplier, threshold)


  def colormap_torch(self, depth, near=0.1):
    depth = depth.clone()
    depth[depth == 0] = torch.inf

    min_depth = torch.clamp(depth, min=near).min()

    inv_depth =  (min_depth / depth).clamp(0, 1)
    inv_depth = (255 * inv_depth).to(torch.int)

    return (self.color_map[inv_depth])

  def colormap_np(self, depth, near=0.1):
    min_depth = np.clip(depth, a_min=near).min()

    inv_depth =  (min_depth / depth)
    inv_depth = (255 * inv_depth).astype(np.uint8)
    return cv2.applyColorMap(inv_depth, cv2.COLORMAP_TURBO)

  def render(self, camera, settings:Settings):
    show = settings.show

    with torch.inference_mode():      
      self.rendering = self.render_gaussians(camera, settings)   
    
    min_depth = self.workspace.camera_extent / 10.

    if settings.view_mode == ViewMode.Depth:
      image_gaussian = self.colormap_torch(self.rendering.depth, 
                  near = min_depth).to(torch.uint8).cpu().numpy()
    else:
      image_gaussian = (self.rendering.image.clamp(0, 1) * 255).to(torch.uint8).cpu().numpy()

    if any([show.initial_points, show.cameras]):

      image, depth = self.pyrender_scene.render(camera, settings)
      depth_gaussian = self.rendering.depth.cpu().numpy()
      depth_gaussian[depth_gaussian == 0] = np.inf

      mask = np.bitwise_and(depth_gaussian > depth, depth > 0)
      return np.where(np.expand_dims(mask, [-1]), image, image_gaussian)
    
    else:
      return image_gaussian




          



    





from dataclasses import dataclass, replace
from beartype import beartype
import torch
from splat_viewer.camera.fov import FOVCamera

from taichi_splatting import renderer
from taichi_splatting.perspective import CameraParams

from splat_viewer.gaussians.data_types import Gaussians, Rendering


def to_camera_params(camera:FOVCamera, device=torch.device("cuda:0")):

  params = CameraParams(
    T_camera_world=torch.from_numpy(camera.camera_t_world),
    T_image_camera=torch.from_numpy(camera.intrinsic),
    
    image_size=tuple(camera.image_size),
    near_plane=camera.near,
    far_plane=camera.far
  )

  return params.to(device=device, dtype=torch.float32)
    

@dataclass
class PackedGaussians:
  gaussians : torch.Tensor
  features : torch.Tensor

  def requires_grad_(self, requires_grad):
    self.gaussians.requires_grad_(requires_grad)
    self.features.requires_grad_(requires_grad)
    return self

class GaussianRenderer:
  @dataclass 
  class Config:
    tile_size : int = 16
    tight_culling : bool = True
    use_depth16 : bool = False

  def __init__(self, **kwargs):
    self.config = GaussianRenderer.Config(**kwargs)

  @beartype
  def pack_inputs(self, gaussians:Gaussians, requires_grad=False):
      packed = torch.cat(
        [gaussians.position, 
         gaussians.log_scaling, 
         gaussians.rotation, 
         gaussians.alpha_logit], 
         dim=-1)
      
      return PackedGaussians(
        gaussians=packed, 
        features=gaussians.sh_feature).requires_grad_(requires_grad)

  
  def update_settings(self, **kwargs):
    self.config = replace(self.config, **kwargs)

  @beartype
  def render(self, inputs:PackedGaussians, camera:FOVCamera, render_depth:bool = True):
    device = inputs.gaussians.device

    config = renderer.RasterConfig(
      tile_size=self.config.tile_size,
      tight_culling=self.config.tight_culling,
    )

    
    rendering = renderer.render_gaussians(
      inputs.gaussians, inputs.features,
      to_camera_params(camera, device),
      config, 
      use_sh=True, 
      use_depth16=self.config.use_depth16,
      render_depth=render_depth)
    
    
    return Rendering(image=rendering.image, 
                            depth=rendering.depth, 
                            depth_var=rendering.depth_var, 
                             camera=camera)
    



from dataclasses import asdict, dataclass, replace
import math
import torch
import torch.nn.functional as F

from beartype import beartype
from splat_viewer.camera.fov import FOVCamera

from splat_viewer.gaussians.data_types import Gaussians, Rendering
from diff_gaussian_rasterization import GaussianRasterizationSettings, GaussianRasterizer

def raster_settings(camera:FOVCamera, sh_degree=3, device=torch.device("cuda:0")):

    proj = camera.ndc_t_camera @ camera.camera_t_world
    view, proj, pos = [torch.from_numpy(t).to(device=device, dtype=torch.float32) 
            for t in (camera.camera_t_world, proj, camera.position)]

    width, height = camera.image_size
    fovW, fovH = camera.fov

    return GaussianRasterizationSettings(
        image_height=int(height),
        image_width=int(width),

        tanfovx=math.tan(fovW * 0.5),
        tanfovy=math.tan(fovH * 0.5),
        bg=torch.tensor([0.0, 0.0, 0.0], device=device),
        scale_modifier=1.0,
        viewmatrix=view.transpose(0, 1),
        projmatrix=proj.transpose(0, 1),
        sh_degree=sh_degree,
        campos=pos,
        prefiltered=False,
        debug=False
    )

  
@dataclass
class PackedGaussians:
  means2D : torch.Tensor
  means3D : torch.Tensor

  shs : torch.Tensor
  alphas : torch.Tensor
  scales : torch.Tensor
  rotations : torch.Tensor

  def requires_grad_(self, requires_grad):
    d = {k:t.requires_grad_(requires_grad) for k, t in asdict(self).items()}
      
    return PackedGaussians(**d)

class DiffGaussianRenderer:

  def __init__(self):
    pass

  @beartype
  def pack_inputs(self, gaussians:Gaussians, requires_grad=False):
      means2D = torch.zeros_like(gaussians.position, device=gaussians.device)

      return PackedGaussians(
        means2D=means2D,
        means3D=gaussians.position,
        shs=gaussians.sh_feature.transpose(1, 2).contiguous(),
        alphas=gaussians.alpha(),
        scales=gaussians.scale(),

        # necessary as we store rotation in xyzw - this uses wxyz
        rotations=torch.roll(F.normalize(gaussians.rotation, dim=1), 1, 1)
      ).requires_grad_(requires_grad)

  
  def update_settings(self, **kwargs):
    pass

  @beartype
  def render(self, inputs:PackedGaussians, camera:FOVCamera, render_depth:bool = True):
    
    sh_degree = int(math.sqrt(inputs.shs.shape[1])) - 1
    settings = raster_settings(camera, sh_degree=sh_degree, device=inputs.means2D.device)
    rasterizer = GaussianRasterizer(raster_settings=settings)

    image, radii = rasterizer.forward(
        means2D=inputs.means2D, means3D=inputs.means3D, shs=inputs.shs, 
        opacities=inputs.alphas, scales=inputs.scales, rotations=inputs.rotations)

    return Rendering(
      image=image.permute(1, 2, 0),
      camera=camera)




from dataclasses import dataclass
from enum import Enum
from typing import Tuple


class ViewMode(Enum):
  Normal = 0
  Depth = 1
  Points = 2
  Hidden = 3

@dataclass(frozen=True)
class Show:
  initial_points: bool = False
  cameras: bool = False
  cropped : bool = False




@dataclass(frozen=True)
class Settings:
  update_rate : int = 20
  move_speed : float = 1.0

  transition_time : float = 0.5
  rotate_speed : float = 1.0

  animate_speed : float = 0.5
  animate_pausing: float = 0.4

  zoom_discrete : float = 1.2
  zoom_continuous : float = 0.1

  drag_speed : float = 1.0
  point_size : float = 2.0

  snapshot_size: Tuple[int, int] = (8192, 6144) 
  snapshot_tile: int = 1024


  device : str = 'cuda:0'
  
  bg_color : Tuple[float, float, float] = (1, 1, 1)

  show : Show = Show()
  view_mode : ViewMode = ViewMode.Normal
  





from PySide6 import QtGui
from PySide6.QtCore import Qt, QEvent

import numpy as np

from ..interaction import Interaction


class FlyControl(Interaction):
  def __init__(self):
    super(FlyControl, self).__init__()
    self.drag_mouse_pos = None

    self.directions = { 
      Qt.Key_Q : np.array([0.,  -1.,  0.]),
      Qt.Key_E : np.array([0.,  1.,  0.]),
      
      Qt.Key_W : np.array([0.,  0.,  1.]),
      Qt.Key_S : np.array([0.,  0.,  -1.]),

      Qt.Key_A : np.array([-1., 0.,  0.]),
      Qt.Key_D : np.array([1.,  0.,  0.])
    }

    self.rotations = { 
      Qt.Key_Z : np.array([0.,  0.,  1.]),
      Qt.Key_C : np.array([0.,  0.,  -1.]),

      Qt.Key_Up : np.array([0.,  1.,  0.]),
      Qt.Key_Down : np.array([0.,  -1.,  0.]),
      
      Qt.Key_Left : np.array([-1.,  0.,  0.]),
      Qt.Key_Right : np.array([1.,  0.,  0.]),
    }


    self.speed_controls = { 
      Qt.Key_Plus : 2.0,
      Qt.Key_Minus : 0.5,
    }


    self.held_keys = set(self.directions.keys()) | set(self.rotations.keys())


  def keyPressEvent(self, event: QtGui.QKeyEvent):
    if event.key() in self.held_keys and not event.isAutoRepeat():
      self.transition(None)

    if event.key() in self.speed_controls and event.modifiers() & Qt.KeypadModifier:
      self.update_setting(move_speed = self.settings.move_speed * self.speed_controls[event.key()])
      return True
      

  def update(self, dt:float):
    scene = self.scene_widget
    mod = 0.1 if Qt.Key_Shift in self.keys_down else 1.0

    for key in self.keys_down:
      if key in self.rotations:
        scene.rotate_camera(mod * self.rotations[key] * dt * self.settings.rotate_speed)

      elif key in self.directions:
        scene.move_camera(mod * self.directions[key] * dt * self.settings.move_speed)
    

  def mousePressEvent(self, event: QtGui.QMouseEvent):
    if event.buttons() & Qt.RightButton:
      self.drag_mouse_pos = event.localPos()
      return True
    

  def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
    if event.button() & Qt.RightButton:
      self.drag_mouse_pos = None
      return True
    
  def mouseMoveEvent(self, event: QtGui.QMouseEvent):
    if event.buttons() & Qt.RightButton and self.drag_mouse_pos is not None:
      delta = event.localPos() - self.drag_mouse_pos

      sz = self.scene_widget.size()

      speed = self.settings.drag_speed
      self.scene_widget.rotate_camera([delta.x() / sz.width() * speed, 
                                -delta.y() / sz.height() * speed,
                                0])
      
      self.drag_mouse_pos = event.localPos()
    



import math
from typing import Callable, List, Optional
from PySide6 import QtGui
from PySide6.QtCore import Qt
from beartype import beartype

import numpy as np
import scipy

from splat_viewer.camera.fov import split_rt

from ..interaction import Interaction
from scipy.spatial.transform import Rotation as R
from scipy.spatial.transform import Slerp



def generalized_sigmoid(x, smooth, eps=1e-6):
    if x < eps:
      return 0.0
    if (1 - x) < eps:
      return 1.0
    else:
      return 1/(1 + (x / (1 - x)) ** -smooth)
      

def animate_to(state, current_pose, dest_pose, on_finish=None):
  to = AnimateCamera([current_pose, dest_pose], loop=False, on_finish=on_finish)
  state.transition(to)

def animate_to_loop(state, current_pose, loop_poses, on_finish=None):
  loop = AnimateCamera(loop_poses, loop=True, on_finish=on_finish)
  to = AnimateCamera([current_pose, loop_poses[0]], loop=False, 
                     on_finish=lambda: state.transition(loop))

  state.transition(to)

class AnimateCamera(Interaction):
  @beartype
  def __init__(self, motion_path:List[np.ndarray], loop=True, 
               on_finish:Optional[Callable]=None):
    super(AnimateCamera, self).__init__()

    bc_type = 'not-a-knot' 

    if loop:
      motion_path = [*motion_path, motion_path[0]]
      bc_type = 'periodic' 

    self.loop = loop
    self.on_finish = on_finish

    self.total = len(motion_path) - 1
    times = np.arange(len(motion_path))    
      
    r, t = zip(*[split_rt(m) for m in motion_path])

    self.rots, self.pos = np.array(r), np.array(t)

    self.slerp = Slerp(times, R.from_matrix(self.rots))
    self.interp = scipy.interpolate.CubicSpline(times, self.pos, axis=0, bc_type=bc_type)
    
    self.t = 0.0  

    self.speed_controls = { 
      Qt.Key_Plus : 1,
      Qt.Key_Minus : -1,
    }


  def keyPressEvent(self, event: QtGui.QKeyEvent):
      if event.key() in self.speed_controls:
        modifier = self.speed_controls[event.key()]

        if event.modifiers() & Qt.ShiftModifier:
          self.update_setting(animate_pausing = np.clip(self.settings.animate_pausing + (0.2 * modifier), 0, 2))
        else:
          self.update_setting(animate_speed = self.settings.animate_speed * (2 ** modifier))

        return True



  def update(self, dt:float):
    scene = self.scene_widget

    inc = dt * self.settings.animate_speed

    if self.t + inc >= self.total:
      if self.on_finish is not None:
        self.on_finish()

    if not self.loop:
      self.t = min(self.total, self.t + inc)
    else:
      self.t = (self.t + inc) % self.total

    frac = math.fmod(self.t, 1)    
    t = math.floor(self.t) + generalized_sigmoid(frac, self.settings.animate_pausing + 1)   

    r = self.slerp(np.array([t])).as_matrix()[0]

    pos = self.interp(t)

    scene.set_camera_pose(r, pos)

    

    





def add_render_arguments(parser):
  parser.add_argument("--tile_size", type=int, default=16, help="tile size for rasterizer")
  parser.add_argument("--no_tight_culling", action="store_true", help="disable tight (OBB) culling")
  parser.add_argument("--depth16", action="store_true", help="use 16 bit depth in sorting (default is 32 bit)")

  parser.add_argument("--taichi", action="store_true", help="use taichi renderer")
  parser.add_argument("--diff_gaussian", action="store_true", help="use diff gaussian renderer")

  return parser


def renderer_from_args(args):
  if args.taichi:
    from splat_viewer.renderer.taichi_3d_gaussian_splatting import TaichiRenderer
    return TaichiRenderer()
  if args.diff_gaussian:
    from splat_viewer.renderer.diff_gaussian_rasterization import DiffGaussianRenderer
    return DiffGaussianRenderer()
  else:
    from splat_viewer.renderer.taichi_splatting import GaussianRenderer
    return GaussianRenderer(tile_size=args.tile_size, 
                                 tight_culling=not args.no_tight_culling,
                                 use_depth16=args.depth16) 



from dataclasses import dataclass, field, replace
from numbers import Number

from pathlib import Path

import numpy as np
import json

from beartype import beartype
from beartype.typing import Tuple

from splat_viewer.camera.transforms import project_points, unproject_pixels


num_pair = np.ndarray | Tuple[Number, Number]
int_pair = np.ndarray | Tuple[int, int]

Box = Tuple[int, int, int, int]


@beartype
@dataclass
class FOVCamera:
   
  position: np.ndarray
  rotation: np.ndarray
  focal_length : np.ndarray # 2
  image_size : np.ndarray # 2

  image_name: str
  principal_point : np.ndarray = field(default_factory=lambda: np.array([0., 0.]))

  near:float  = 0.1
  far :float  = 50.0

  
  @property
  def aspect(self):
    width, height = self.image_size
    return width / height
  
  @property
  def width(self):
    return self.image_size[0]
  
  @property
  def height(self):
    return self.image_size[1]
  
  def scale_size(self, scale_factor) -> 'FOVCamera':

    return replace(self,
      image_size=np.round(self.image_size * scale_factor).astype(np.int32),
      focal_length=self.focal_length * scale_factor,
      principal_point=self.principal_point * scale_factor
    )
    
  def crop_offset_size(self, offset:num_pair, size:num_pair) -> 'FOVCamera':
    offset, size = np.array(offset), np.array(size)
    return replace(self,
      image_size=size.astype(np.int32),
      principal_point=self.principal_point - offset
    )
  
  def crop_extent(self, centre:num_pair, size:num_pair) -> 'FOVCamera':
      centre, size = np.array(centre), np.array(size)
      return self.crop_offset_size(centre - size / 2, size)

  def crop_box(self, box:Box) -> 'FOVCamera':
    x_min, x_max, y_min, y_max = box
    return self.crop_offset_size(
      np.array([x_min, y_min]),
      np.array([x_max - x_min, y_max - y_min])
    )
    
  def pad_to(self, image_size:num_pair) -> 'FOVCamera':
    image_size = np.array(image_size)
    return replace(self,
      image_size=image_size.astype(np.int32),
      principal_point=self.principal_point + (image_size - self.image_size) / 2
    )
    

  def resize_shortest(self, min_size, max_size) -> 'FOVCamera':
    shortest = min(self.image_size)
    scale = (min_size / shortest if shortest < min_size 
             else max_size / shortest)
    return self.scale_size(scale)
  
  def resize_longest(self, size) -> 'FOVCamera':
    longest = max(self.image_size)
    return self.scale_size(size / longest)

  
  def resize_to(self, size:num_pair) -> 'FOVCamera':
    size = np.array(size)
    return self.scale_size(size / self.image_size)
  

  def zoom(self, zoom_factor) -> 'FOVCamera':
    return replace(self, focal_length=self.focal_length * zoom_factor)


  @property
  def world_t_camera(self):
    return join_rt(self.rotation, self.position)
  
  @property
  def camera_t_world(self):
    return np.linalg.inv(self.world_t_camera)
  
  def __repr__(self):
    w, h = self.image_size
    fx, fy = self.focal_length
    cx, cy = self.principal_point
    return f"FOVCamera(name={self.image_name}@{w}x{h} pos={self.position}, z={self.forward}, fx={fx} fy={fy}, cx={cx} cy={cy})"

  def __str__(self):
    return repr(self)

  
  @property
  def right(self):
    return self.rotation[0]

  @property
  def up(self):
    return -self.rotation[1]

  @property
  def forward(self):
    return self.rotation[2]
  

  @property
  def fov(self):
    return np.arctan2(self.image_size, self.focal_length * 2) * 2 
  
  @property
  def intrinsic(self):
  
    cx, cy = self.principal_point
    fx, fy = self.focal_length

    return np.array(
      [[fx, 0,  cx],
        [0, fy, cy],
        [0, 0,  1]]
    )
  
  def unproject_pixels(self, xy:np.ndarray, depth:np.ndarray):
     return unproject_pixels(self.world_t_image, xy, depth)

  def project_points(self, points:np.ndarray):
    return project_points(self.image_t_world, points)
    

  def unproject_pixel(self, x, y, depth):
    points = self.unproject_pixels(np.array([[x, y]]), np.array([[depth]]))
    return tuple(points[0])
  
  def project_point(self, x, y, z):
    xy, depth = self.project_points(np.array([[x, y, z]]))
    return tuple([*xy[0], *depth[0]])

  @property
  def image_t_camera(self):
    m44 = np.eye(4)
    m44[:3, :3] = self.intrinsic
    
    return m44
  
  @property
  def image_t_world(self):
    return self.image_t_camera @ self.camera_t_world
  
  @property
  def world_t_image(self):
    return np.linalg.inv(self.image_t_world)

  @property
  def projection(self):
    return self.image_t_world
  
  @property
  def ndc_t_camera(self):
    """ OpenGL projection - Camera to Normalised Device Coordinates (NDC)
    """
    w, h = self.image_size

    cx, cy = self.principal_point
    fx, fy = self.focal_length
    n, f = self.near, self.far

    return np.array([
          [2.0 * fx / w,   0,              1.0 - 2.0 * cx / w,   0],
          [0,              2.0 * fy / h,   2.0 * cy / h - 1.0,   0],
          [0,              0,              (f + n) / (n - f),    (2 * f * n) / (n - f)],
          [0,              0,              1.0,                 0]
      ], dtype=np.float32)
  


  @property
  def gl_camera_t_image(self):
    return np.linalg.inv(self.ndc_t_camera)

  @property
  def gl_camera_t_world(self):

    flip_yz = np.array([
      [1, 0, 0],
      [0, -1, 0],
      [0, 0, -1]
    ])

    rotation = self.rotation  @ flip_yz 
    return join_rt(rotation, self.position)

  @property
  def ndc_t_world(self):

    return self.ndc_t_camera @ self.gl_camera_t_world


def join_rt(R, T):
  Rt = np.zeros((4, 4))
  Rt[:3, :3] = R
  Rt[:3, 3] = T
  Rt[3, 3] = 1.0

  return Rt
         

def split_rt(Rt):
  R = Rt[:3, :3]
  T = Rt[:3, 3]
  return R, T



def from_json(camera_info) -> FOVCamera:
  pos = np.array(camera_info['position'])
  rotation = np.array(camera_info['rotation']).reshape(3, 3)
  w, h = (camera_info['width'], camera_info['height'])
  cx, cy = (camera_info.get('cx', w/2.), camera_info.get('cy', h/2.))
  

  return FOVCamera(
    position=pos,
    rotation=rotation,
    image_size=np.array([w, h], dtype=np.int32),
    focal_length=np.array([camera_info['fx'], camera_info['fy']]),
    principal_point=np.array([cx, cy]),
    image_name=camera_info['img_name']
  )



def load_camera_json(filename:Path):
  cameras = sorted(json.loads(filename.read_text()), key=lambda x: x['id'])

  return {camera_info['id']: from_json(camera_info) for camera_info in cameras}
  




import argparse
from natsort import natsorted
from pathlib import Path
from typing import List
import shutil


def main():
  parser = argparse.ArgumentParser(description="Export a trained workspace (only the minimum files required)")
  parser.add_argument("model_path", type=Path)
  parser.add_argument("output", type=Path)

  args = parser.parse_args()

  clouds = natsorted([file for file in args.model_path.glob("point_cloud/iteration_*/point_cloud.ply")])

  if len(clouds) == 0:
    raise Exception("No point clouds found in {}".format(args.model_path))
  
  cloud_file = clouds[-1]
  print("Using point cloud {}".format(cloud_file))

  camera_file = args.model_path/"cameras.json"
  scene_file = args.model_path/"scene.json"
  input_file = args.model_path/"input.ply"
  cfg_file = args.model_path/"cfg_args"
  
  files:List[Path] = [camera_file, input_file, cloud_file, cfg_file, scene_file]
  
  for file in files:
    if not file.exists():
      raise Exception(f"File not found {file}")

  
  args.output.mkdir(parents=True, exist_ok=True)
  for file in files:
    filename = file.relative_to(args.model_path)
    out_filename = args.output/filename

    print(f"Copying {file} to {out_filename}")
    out_filename.parent.mkdir(parents=True, exist_ok=True)

    shutil.copyfile(file, out_filename)







if __name__=="__main__":
  main()



import sys
import argparse

from PySide6 import  QtWidgets
from PySide6.QtWidgets import QApplication

from splat_viewer.gaussians.workspace import Workspace, load_workspace
from splat_viewer.gaussians import Gaussians
from splat_viewer.renderer.arguments import add_render_arguments, renderer_from_args
from splat_viewer.renderer.taichi_splatting import GaussianRenderer

from splat_viewer.viewer.scene_widget import SceneWidget, Settings

import signal
import taichi as ti
import torch


def show_workspace(workspace:Workspace, gaussians:Gaussians = None):
  import taichi as ti
  ti.init(ti.gpu, offline_cache=True, device_memory_GB=0.1)

  from splat_viewer.viewer.viewer import sigint_handler
  signal.signal(signal.SIGINT, sigint_handler)

  
  app = QtWidgets.QApplication(["viewer"])
  widget = SceneWidget(renderer = GaussianRenderer())

  if gaussians is None:
    gaussians = workspace.load_model(workspace.latest_iteration())

  print(f"Showing model from {workspace.model_path}: {gaussians}")


  widget.load_workspace(workspace, gaussians)
  widget.show()
  app.exec_()



def process_cl_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('model_path', help="workspace folder containing cameras.json, input.ply and point_cloud folder with .ply models")  # positional argument
    parser.add_argument('--model', default=None, help="load model from point_clouds folder, default is latest iteration") 
    parser.add_argument('--device', default='cuda:0', help="torch device to use")
    parser.add_argument('--debug', action='store_true', help="enable taichi kernels in debug mode")

    add_render_arguments(parser)    

    parsed_args, unparsed_args = parser.parse_known_args()
    return parsed_args, unparsed_args

def sigint_handler(*args):
    QApplication.quit()


def main():
    signal.signal(signal.SIGINT, sigint_handler)
    torch.set_printoptions(precision=5, sci_mode=False, linewidth=120)


    parsed_args, unparsed_args = process_cl_args()
    workspace = load_workspace(parsed_args.model_path)

    if parsed_args.model is None:
      parsed_args.model = workspace.latest_iteration()

    gaussians = workspace.load_model(parsed_args.model)
    print(f"Loaded model {parsed_args.model}: {gaussians}")

    ti.init(ti.gpu, offline_cache=True, debug=parsed_args.debug, device_memory_GB=0.1)


    qt_args = sys.argv[:1] + unparsed_args
    app = QApplication(qt_args)
    

    window = QtWidgets.QMainWindow()

    scene_widget = SceneWidget(
       settings=Settings(device=parsed_args.device),
       renderer = renderer_from_args(parsed_args)
    )
    
    scene_widget.load_workspace(workspace, gaussians)

    window.setCentralWidget(scene_widget)

    window.show()
    sys.exit(app.exec())

  
if __name__ == '__main__':
  main()


from PySide6.QtCore import Qt, QEvent


keymap = {}
for key, value in vars(Qt).items():
    if isinstance(value, Qt.Key):
        keymap[value] = key

def keyevent_to_string(event):
    sequence = []

    key = keymap.get(event.key(), event.text())
    if key not in sequence:
        sequence.append(key)
    return '+'.join(sequence)